package de.uni_passau.fim.seibt.gitwrapper.repo;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.Scanner;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.ExecRes;
import de.uni_passau.fim.seibt.gitwrapper.process.ToolNotWorkingException;
import de.uni_passau.fim.seibt.gitwrapper.process.ToolWrapper;

import static de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.asList;

/**
 * A wrapper for executing git commands.
 */
public class GitWrapper extends ToolWrapper {

    private static final Logger LOG = Logger.getLogger(GitWrapper.class.getCanonicalName());

    private static final String FATAL_PREFIX = "fatal";
    private static final String ERROR_PREFIX = "error";

    private static final Pattern CLONING_INTO = Pattern.compile("Cloning into '(.*)'\\.\\.\\.");
    private static final Pattern ALREADY_EXISTS = Pattern.compile("fatal: destination path '(.*)' already exists and is not an empty directory\\.");

    private static final String GIT_VERSION_C = "--version";
    private static final String GIT_VERSION_PREFIX = "git version";

    private static final Map<String, String> env;

    static {
        env = new HashMap<>();

        String locale = "C";
        env.put("LANGUAGE", locale);
        env.put("LC_ALL", locale);
        env.put("LANG", locale);
    }

    /**
     * Constructs a new {@link GitWrapper}
     *
     * @param git
     *         the git command, e.g. '/usr/bin/git' or just 'git'
     */
    public GitWrapper(String git) throws ToolNotWorkingException {
        super(git);
    }

    @Override
    protected boolean isWorking() {
        Function<ExecRes, Boolean> checkPrefix = execRes -> execRes.getStdOutTrimmed().startsWith(GIT_VERSION_PREFIX);
        return exec(new File("."), true, asList(GIT_VERSION_C)).map(checkPrefix).orElse(false);
    }

    /**
     * Attempts to clone the git repository at the given <code>url</code> into a directory under <code>parentDir</code>.
     * If <code>parentDir</code> already contains a git repository with the name that would be produced when cloning
     * <code>url</code> it is assumed to be a clone of <code>url</code> and returned as a {@link Repository}.
     *
     * @param parentDir
     *         the directory in which to place the clone of <code>url</code>
     * @param url
     *         the git repository to clone
     * @param fetch
     *         whether to perform a fetch of the {@link Repository} if a clone of it was already found on disk
     * @return the resulting {@link Repository} or an empty {@link Optional} if there is an error
     */
    public Optional<Repository> clone(File parentDir, String url, boolean fetch) {
        LOG.fine(() -> String.format("Cloning '%s'.", url));

        Optional<ExecRes> res = exec(parentDir, "clone", url);
        Function<ExecRes, Repository> toRepo = execRes -> {
            Scanner sc = new Scanner(execRes.getStdOutTrimmed());

            if (!sc.hasNextLine()) {
                LOG.warning(() -> String.format("No output while cloning '%s'.", url));
                return null;
            }

            String firstLine = sc.nextLine();
            Matcher matcher;

            if (execRes.exitCode == EXIT_SUCCESS) {
                matcher = CLONING_INTO.matcher(firstLine);
            } else {
                matcher = ALREADY_EXISTS.matcher(firstLine);
            }

            if (!matcher.find()) {
                LOG.warning(() -> String.format("Unexpected output while cloning '%s'.", url));
                return null;
            }

            String name = matcher.group(1);
            File repoDir = new File(parentDir, name);
            String repoPath = repoDir.getAbsolutePath();

            if (execRes.exitCode == EXIT_SUCCESS) {
                LOG.fine(() -> String.format("Cloned '%s' into '%s'.", url, repoPath));
                return new Repository(this, url, repoDir);
            } else {
                LOG.fine(() -> String.format("Found '%s' which would be produced for '%s'", repoPath, url));

                if (!isGitDir(repoDir)) {
                    LOG.warning(() -> String.format("'%s' is not a git repository.", repoPath));
                    return null;
                }

                LOG.fine(() -> String.format("'%s' is a git repository. Assuming it is a clone of '%s'.", repoPath, url));
                Repository repo = new Repository(this, url, repoDir);

                if (fetch && !repo.fetch()) {
                    LOG.warning("Could not fetch " + repo + ". It may be out of date.");
                }

                return repo;
            }
        };

        return res.map(toRepo);
    }

    /**
     * Imports an existing repository.
     *
     * @param directory
     *         the path to the repository
     * @return the {@link Repository} if the import was successful or an empty {@link Optional} if there was an error
     */
    public Optional<Repository> importRepository(File directory) {
        if (!isGitDir(directory)) {
            LOG.warning(() -> directory + " does not contain a git repository.");
            return Optional.empty();
        }

        Optional<ExecRes> originURL = exec(directory, "config", "--get", "remote.origin.url");
        Function<ExecRes, Repository> toRepository = res -> {
            String repoUrl = null;

            if (failed(res)) {
                LOG.warning(() -> String.format("Failed to get the origin url for the repository in %s. Using the " +
                                                "canonical directory path.", directory));
            } else if (res.getStdOutTrimmed().isEmpty()) {
                LOG.warning(() -> String.format("Repository in %s does not have an origin url. Using the canonical " +
                                                "directory path.", directory));
            } else {
                repoUrl = res.getStdOutTrimmed();
            }

            if (repoUrl == null) {

                try {
                    repoUrl = directory.getCanonicalPath();
                } catch (IOException e) {
                    LOG.log(Level.WARNING, e, () -> "Exception computing the canonical path for " + directory +
                                                    ". Using the absolute path instead.");
                    repoUrl = directory.getAbsolutePath();
                }
            }

            return new Repository(this, repoUrl, directory);
        };

        return originURL.map(toRepository);
    }

    /**
     * Returns whether the given directory (or any of its parent directories) is a git repository. Returns false if
     * <code>directory</code> is not a directory.
     *
     * @param directory
     *         the directory to check
     * @return true iff the given directory (or any of its parent directories) is a git repository
     */
    public boolean isGitDir(File directory) {

        if (!directory.isDirectory()) {
            LOG.warning(() -> String.format("The file '%s' does not exist or is not a directory.", directory));
            return false;
        }

        Optional<ExecRes> status = exec(directory, "rev-parse", "--is-inside-git-dir");
        return status.map(res -> !failed(res)).orElse(false);
    }

    /**
     * Returns whether the given git command failed. This relies on exit code first and then on the assumption that
     * the output (in case of failure) starts with either "{@value FATAL_PREFIX}" or "{@value ERROR_PREFIX}".
     *
     * @param res
     *         the {@link ExecRes} to check
     * @return whether the git command failed
     */
    public boolean failed(ExecRes res) {
        return res.exitCode != EXIT_SUCCESS || res.getStdOutTrimmed().startsWith(FATAL_PREFIX)
                                            || res.getStdOutTrimmed().startsWith(ERROR_PREFIX);
    }

    /**
     * Executes '&lt;git command&gt &lt;parameters&gt' and returns the exit code and the output of the command. If there is an
     * exception executing the command an empty {@link Optional} will be returned.
     *
     * @param workingDirectory
     *         the working directory for the command execution
     * @param parameters
     *         the git parameters
     * @return the result of the execution
     */
    public Optional<ExecRes> exec(File workingDirectory, String... parameters) {
        return exec(workingDirectory, true, env, asList(parameters));
    }
}
